/******************************************************************************
 * Copyright Codito Technologies (www.codito.com) Oct 01, 2004
 * 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *****************************************************************************/

/*
 * drivers/char/arcserial.c
 *
 * (C) 2003, ARC International

 * Driver for console on serial interface for the ArcAngel3 board. We
 * use VUART0 as the console device.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/console.h>
#include <linux/serial.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/slab.h>
#include <linux/interrupt.h>

#include <asm/irq.h>
#include <asm/page.h>
#include <asm/uaccess.h>
#include <asm/serial.h>

#ifdef CONFIG_SKY_IR
#include <asm/io.h>
#include <linux/ioport.h>
#include "sky_ir.h"	
#endif


#undef SERIAL_PARANOIA_CHECK

/* Need to know the system clock frequency for timing set-up */
extern unsigned long arc700_clkfreq;


int last_serial_int = -1;

static struct aa3_serial aa3_info[NR_PORTS];
struct aa3_serial *IRQ_ports[NR_IRQS];

static unsigned int uart_irqs[NR_PORTS] = UART_IRQ_DEFNS;

aa3_uart *uart_addr[NR_PORTS] = UART_BASEADDR;

int console_inited = 0; /* have we initialized the console already? */

struct tty_driver serial_driver, callout_driver;
static int serial_refcount;

static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

#define MAX_NAMED_UART 4
static const char* uart_names[] = { "UART 0", "UART 1", "UART 2", "UART 3" };

DECLARE_TASK_QUEUE(tq_serial);

/*
 * tmp_buf is used as a temporary buffer by serial_write. We need to
 * lock it in case the memcpy_fromfs blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.
 */
static unsigned char tmp_buf[SERIAL_XMIT_SIZE]; /* This is cheating */
DECLARE_MUTEX(tmp_buf_sem);

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

static inline int serial_paranoia_check(struct aa3_serial *info,
					kdev_t device, const char *routine)
{
#ifdef SERIAL_PARANOIA_CHECK
	static const char *badmagic =
		"Warning: bad magic number for serial struct (%s) in %s\n";
	static const char *badinfo =
		"Warning: null async_struct for (%s) in %s\n";

	if (!info) {
		printk(badinfo, kdevname(device), routine);
		return 1;
	}
	if (info->magic != SERIAL_MAGIC) {
		printk(badmagic, kdevname(device), routine);
		return 1;
	}
#endif
	return 0;
}

/*
 * ------------------------------------------------------------
 * arcserial_stop() and arcserial_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 * ------------------------------------------------------------
 */

static void arcserial_stop(struct tty_struct *tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;
	aa3_uart *uart = uart_addr[info->line];
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "arcserial_stop"))
		return;
	
	save_flags(flags); cli();
	uart->status &= ~UART_TXENB;
	restore_flags(flags);
}

/*
 * Output a single character, using UART polled mode.
 * This is used for console output.
 */

void arcserial_put_char(char ch)
{

        aa3_uart *uart = uart_addr[0];

	/* check if TXEMPTY (bit 7) in the status reg. is 1. Else, we
	 * aren't supposed to write right now. Wait for some time.
	 */

	while(!(uart->status & UART_TXEMPTY));

	/* actually write the character */
	uart->data = ch;
}

static void arcserial_start(struct tty_struct *tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;
	unsigned long flags;
	aa3_uart *uart = uart_addr[info->line];
	
	if (serial_paranoia_check(info, tty->device, "arcserial_start"))
		return;
	
	save_flags(flags); cli();
	if (info->xmit_cnt && info->xmit_buf && !(uart->status & UART_TXENB))
		uart->status |= UART_TXENB;
	restore_flags(flags);
}


/*
 * This routine is used by the interrupt handler to schedule
 * processing in the software interrupt portion of the driver.
 */
static inline void arcserial_sched_event(struct aa3_serial *info,
						int event)
{
	info->event |= 1 << event;
	queue_task(&info->tqueue, &tq_serial);
	mark_bh(SERIAL_BH);
}

static void receive_chars(struct aa3_serial *info, struct pt_regs *regs, unsigned short rx)
{
	struct tty_struct *tty = info->tty;
	unsigned char ch;
	aa3_uart *uart = uart_addr[info->line];
	int status;
	
	if(!tty)
	{
		printk("AA3 Serial: missing tty on line %d\n", info->line);
		goto clear_and_exit;
	}
		
	do {
		ch = rx & 0x00ff;

		if(uart->status & UART_OVERFLOW_ERR) {
			*tty->flip.flag_buf_ptr++ = TTY_OVERRUN;
			printk("AA3 Serial Driver: Overflow Error while receiving a character\n");
			uart->status &= ~UART_OVERFLOW_ERR; 
		} else if(uart->status & UART_FRAME_ERR) {
			*tty->flip.flag_buf_ptr++ = TTY_FRAME;
			printk("AA3 Serial Driver: Framing Error while receiving a character\n");
			uart->status &= ~UART_FRAME_ERR;	
		} else {
			*tty->flip.flag_buf_ptr++ = 0; /* XXX */
		}
		*tty->flip.char_buf_ptr++ = ch;
			tty->flip.count++;
		
		/*
		 * If the buffer's full, get out and flush it before we pick up the next character
		 */
		if (tty->flip.count >= TTY_FLIPBUF_SIZE) {
			queue_task(&tty->flip.tqueue, &tq_timer);
			return;
		}

		if(!(status = uart->status & UART_RXEMPTY))
			 rx = uart->data;
		/* keep reading characters till the RxFIFO becomes empty */
	}while(!status);

	queue_task(&tty->flip.tqueue, &tq_timer);

clear_and_exit:
	return;
}

static void transmit_chars(struct aa3_serial *info)
{
	aa3_uart *uart = uart_addr[info->line];

	if (info->x_char) {
		/* Send next char */
		uart->data = info->x_char;
		info->x_char = 0;
		goto clear_and_return;
	}

	if((info->xmit_cnt <= 0) || info->tty->stopped) {
		/* That's peculiar... TX ints off */
		uart->status &= ~UART_TXENB;
		goto clear_and_return;
	}

	/* Send char */
	uart->data = info->xmit_buf[info->xmit_tail++];
	info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE-1);
	info->xmit_cnt--;

	if (info->xmit_cnt < WAKEUP_CHARS)
		arcserial_sched_event(info, ARCSERIAL_EVENT_WRITE_WAKEUP);

	if(info->xmit_cnt <= 0) {
		/* All done for now... TX ints off */
		uart->status &= ~UART_TXENB;
		goto clear_and_return;
	}

clear_and_return:
	/* Clear interrupt (should be auto)*/
	return;
}

#ifdef CONFIG_SKY_IR
void sky_ir_interrupt(int irq, void *dev_id, struct pt_regs *regs){

	// pause key
	if(IR3CD1==0xaa22a888 && IR3CD2==0x8888a888 && IR3CD3==0xaaaa8888 && IR3CD4==0x00000000)
	{
		struct aa3_serial *info;
		aa3_uart *uart;
		
		info = IRQ_ports[VUART_INT];
		if(!info)
				return;

		uart = uart_addr[info->line];
			
		receive_chars(info, regs, 0x20);		
	}

	IR_INT_RESET;	
}
#endif

/*
 * This is the serial driver's generic interrupt routine
 */
void aa3_serial_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct aa3_serial *info;
	aa3_uart *uart;

	info = IRQ_ports[irq];
	if(!info)
			return;
	
	uart = uart_addr[info->line];
	
	if((uart->status & UART_RXENB) && !(uart->status & UART_RXEMPTY))
		receive_chars(info, regs, uart->data);		
	if((uart->status & UART_TXENB) && (uart->status & UART_TXEMPTY))
	{
		last_serial_int = irq;
		transmit_chars(info);
	}
	
	if(uart->status & UART_FRAME_ERR) 
		uart->status &= ~UART_FRAME_ERR;
	if(uart->status & UART_OVERFLOW_ERR)
		uart->status &= ~UART_OVERFLOW_ERR;
 
}

/*
 * This routine is used to handle the "bottom half" processing for the
 * serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * arcserial_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON. This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using arcserial_sched_event(), and they get done here.
 */
static void do_serial_bh(void)
{
	run_task_queue(&tq_serial);
}

static void do_softint(void *private_)
{
	struct aa3_serial *info = (struct aa3_serial *) private_;
	struct tty_struct *tty;
	
	tty = info->tty;
	if (!tty)
		return;

	if (test_and_clear_bit(ARCSERIAL_EVENT_WRITE_WAKEUP, &info->event)) {
		tty_wakeup(tty);
	}
}

/*
 * This routine is called from the scheduler tqueue when the interrupt
 * routine has signalled that a hangup has occurred.	The path of
 * hangup processing is:
 *
 *	serial interrupt routine -> (scheduler tqueue) ->
 *	do_serial_hangup() -> tty->hangup() -> arcserial_hangup()
 * 
 */
static void do_serial_hangup(void *private_)
{
	struct aa3_serial *info = (struct aa3_serial *) private_;
	struct tty_struct *tty;
	
	tty = info->tty;
	if (!tty)
		return;

	tty_hangup(tty);
}

//-- Flowbee 2007-12-11
static void uart_rx_on_off(struct aa3_serial *info, unsigned long onoff)
{
	aa3_uart *uart = uart_addr[info->line];
	unsigned long baud_h, baud_l;
	unsigned short port;
	unsigned long tmp;

	if (!info->tty || !info->tty->termios)
		return;
	if (!(port = info->port))
		return;

//	printk("Changing baud to %d\n",newbaud);
	if(onoff == 1){
		uart->status = 0;
		uart->status |= UART_RXENB;
		//printk("\nUART RXint on!");
	}
	else{
		uart->status &= ~(UART_RXENB);
		//printk("\nUART RXint off!");
	}
}

static void change_speed(struct aa3_serial *info, unsigned long newbaud)
{
	aa3_uart *uart = uart_addr[info->line];
	unsigned long baud_h, baud_l;
	unsigned short port;
	unsigned long tmp;

	if (!info->tty || !info->tty->termios)
		return;
	if (!(port = info->port))
		return;

//	printk("Changing baud to %d\n",newbaud);

	//tmp = uart->data;
	while (! (uart->status & UART_TXEMPTY)) {}

	tmp = (arc700_clkfreq/(newbaud*4)) - 1;

	info->baud = (int) newbaud;
	baud_l = (tmp & 0xff);
	baud_h = (tmp & 0xff00) >> 8;

	uart->baudl = baud_l;
	uart->baudh = baud_h;
}

static int startup(struct aa3_serial * info)
{
	aa3_uart *uart = uart_addr[info->line];
	unsigned long flags;
	
	if (info->flags & ASYNC_INITIALIZED)
		return 0;

	if (!info->xmit_buf) {
		info->xmit_buf = (unsigned char *)kmalloc(PAGE_SIZE, GFP_KERNEL);
		if (!info->xmit_buf)
			return -ENOMEM;
	}

	save_flags(flags); cli();

	/*
	 * Clear the FIFO buffers and disable them
	 * (they will be reenabled in change_speed())
	 */

	info->xmit_fifo_size = 1;
	uart->status = UART_RXENB | UART_TXENB;

	if (info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;

	/*
	 * and set the speed of the serial port
	 */
	change_speed(info, CONSOLE_BAUD_RATE);

	info->flags |= ASYNC_INITIALIZED;
	restore_flags(flags);
	return 0;
}

/*
 * This routine will shutdown a serial port; interrupts are disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 */
static void shutdown(struct aa3_serial * info)
{
	aa3_uart *uart = uart_addr[info->line];
	unsigned long flags;

	uart->status = 0; /* All off! */
	if (!(info->flags & ASYNC_INITIALIZED))
		return;

	save_flags(flags); cli(); /* Disable interrupts */
	
	if (info->xmit_buf) {
		free_page((unsigned long) info->xmit_buf);
		info->xmit_buf = 0;
	}

	if (info->tty)
		set_bit(TTY_IO_ERROR, &info->tty->flags);
	
	info->flags &= ~ASYNC_INITIALIZED;
	restore_flags(flags);
}

/*
static void arcserial_set_ldisc(struct tty_struct *tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;

	if (serial_paranoia_check(info, tty->device, "arcserial_set_ldisc"))
		return;

	info->is_cons = (tty->termios->c_line == N_TTY);
	
	printk("AA3 serial driver: ttyS%d console mode %s\n", info->line, info->is_cons ? "on" : "off");
}
*/

static void arcserial_flush_chars(struct tty_struct *tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;
	aa3_uart *uart = uart_addr[info->line];
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "arcserial_flush_chars"))
		return;

	if (info->xmit_cnt <= 0 || tty->stopped || tty->hw_stopped ||
			!info->xmit_buf) 
		return;

	/* Enable transmitter */
	save_flags(flags); cli();

	uart->status |= UART_TXENB;

	if (uart->status & UART_TXEMPTY) {
		/* Send char */
		uart->data = info->xmit_buf[info->xmit_tail++];
		info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE-1);
		info->xmit_cnt--;
	}
	restore_flags(flags);
}


static int arcserial_write(struct tty_struct * tty, int from_user,
				const unsigned char *buf, int count)
{
	int c, total = 0;
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;
	aa3_uart *uart = uart_addr[info->line];
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "arcserial_write"))
		return 0;

	if (!tty || !info->xmit_buf)
		return 0;

	save_flags(flags);
	while (1) {
		cli();
		c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
					 SERIAL_XMIT_SIZE - info->xmit_head));
		if (c <= 0)
			break;

		if (from_user) {
			down(&tmp_buf_sem);
			copy_from_user(tmp_buf, buf, c);
			c = MIN(c, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
							 SERIAL_XMIT_SIZE - info->xmit_head));
			memcpy(info->xmit_buf + info->xmit_head, tmp_buf, c);
			up(&tmp_buf_sem);
		} else
			memcpy(info->xmit_buf + info->xmit_head, buf, c);
		info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE-1);
		info->xmit_cnt += c;
		restore_flags(flags);
		buf += c;
		count -= c;
		total += c;
	}
	if (info->xmit_cnt && !tty->stopped && !tty->hw_stopped) {
		/* Enable transmitter */
		cli();
				uart->status |= UART_TXENB;
		if (uart->status & UART_TXEMPTY) {
			uart->data = info->xmit_buf[info->xmit_tail++];
			info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE-1);
			info->xmit_cnt--;
		}

		//		restore_flags(flags);
	}
	restore_flags(flags);

	return total;
}

#if 0
static int arcserial_write_room(struct tty_struct *tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;
	volatile int	ret;
				
	if (serial_paranoia_check(info, tty->device, "arcserial_write_room"))
		return 0;
	ret = (SERIAL_XMIT_SIZE - info->xmit_cnt) - 1;
	if (ret < 0)
		return 0;
	else
		return ret;
}
#else
static int arcserial_write_room(struct tty_struct *tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;
	int ret;
				
	if (serial_paranoia_check(info, tty->device, "arcserial_write_room"))
		return 0;
	ret = (SERIAL_XMIT_SIZE - info->xmit_cnt) - 1;
	if (ret < 0)
		return 0;
	else
		return ret;
}
#endif

static int arcserial_chars_in_buffer(struct tty_struct *tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;
				
	if (serial_paranoia_check(info, tty->device, "arcserial_chars_in_buffer"))
		return 0;
	return info->xmit_cnt;
}

static void arcserial_flush_buffer(struct tty_struct *tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;
				
	if (serial_paranoia_check(info, tty->device, "arcserial_flush_buffer"))
		return;
	cli();
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	sti();
	tty_wakeup(tty);
}

/*
 * ------------------------------------------------------------
 * arcserial_throttle()
 * 
 * This routine is called by the upper-layer tty layer to signal that
 * incoming characters should be throttled.
 * ------------------------------------------------------------
 */

static void arcserial_throttle(struct tty_struct * tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;

	if (serial_paranoia_check(info, tty->device, "arcserial_throttle"))
		return;
	
	if (I_IXOFF(tty))
		info->x_char = STOP_CHAR(tty);

	/* Turn off RTS line (do this atomic) */
}

static void arcserial_unthrottle(struct tty_struct * tty)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;

	if (serial_paranoia_check(info, tty->device, "arcserial_unthrottle"))
		return;
	
	if (I_IXOFF(tty)) {
		if (info->x_char)
			info->x_char = 0;
		else
			info->x_char = START_CHAR(tty);
	}

	/* Assert RTS line (do this atomic) */
}

/*
 * ------------------------------------------------------------
 * arcserial_ioctl() and friends
 * ------------------------------------------------------------
 */

// FIXME :: uncomment when ioctl's are implemented 
#if 0

static int get_serial_info(struct aa3_serial * info,
				 struct serial_struct * retinfo)
{
	struct serial_struct tmp;
	
	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = info->type;
	tmp.line = info->line;
	tmp.port = info->port;
	tmp.irq = info->irq;
	tmp.flags = info->flags;
	tmp.baud_base = info->baud_base;
	tmp.close_delay = info->close_delay;
	tmp.closing_wait = info->closing_wait;
	tmp.custom_divisor = info->custom_divisor;
	copy_to_user(retinfo,&tmp,sizeof(*retinfo));
	return 0;
}

static int set_serial_info(struct aa3_serial * info,
				 struct serial_struct * new_info)
{
	struct serial_struct new_serial;
	struct aa3_serial old_info;
	int				retval = 0;

	if (!new_info)
		return -EFAULT;
	copy_from_user(&new_serial,new_info,sizeof(new_serial));
	old_info = *info;

	if (!suser()) {
		if ((new_serial.baud_base != info->baud_base) ||
				(new_serial.type != info->type) ||
				(new_serial.close_delay != info->close_delay) ||
				((new_serial.flags & ~ASYNC_USR_MASK) !=
				 (info->flags & ~ASYNC_USR_MASK)))
			return -EPERM;
		info->flags = ((info->flags & ~ASYNC_USR_MASK) |
						 (new_serial.flags & ASYNC_USR_MASK));
		info->custom_divisor = new_serial.custom_divisor;
		goto check_and_exit;
	}

	if (info->count > 1)
		return -EBUSY;

	/*
	 * OK, past this point, all the error checking has been done.
	 * At this point, we start making changes.....
	 */

	info->baud_base = new_serial.baud_base;
	info->flags = ((info->flags & ~ASYNC_FLAGS) |
			(new_serial.flags & ASYNC_FLAGS));
	info->type = new_serial.type;
	info->close_delay = new_serial.close_delay;
	info->closing_wait = new_serial.closing_wait;

check_and_exit:
	retval = startup(info);
	return retval;
}
#endif	/* if 0 */ 

static int arcserial_ioctl(struct tty_struct *tty, struct file *file, 
				unsigned int cmd, unsigned long arg)
{
//	printk("ARC: aa3 serial_driver, ioctl still to be implemented !!!\n");
	struct aa3_serial * info = (struct aa3_serial *)tty->driver_data;
	aa3_uart *uart = uart_addr[info->line];
	unsigned long tmpul;
//	printk("ARC: aa3 serial_driver, ioctl line %d, cmd 0x%x\n",info->line,cmd);
	
	switch (cmd)
	{
		case GET_BAUD:
			return copy_to_user((void *) arg, &(info->baud), sizeof(unsigned long)) ? -EFAULT : 0;
		 
		case SET_BAUD:
			if (copy_from_user((void *)&tmpul, (void *) arg, sizeof(unsigned long)))
				return -EFAULT;

			change_speed(info, tmpul);
			return 0;

		case UART_RX_ON_OFF:
			if (copy_from_user((void *)&tmpul, (void *) arg, sizeof(unsigned long)))
				return -EFAULT;

			uart_rx_on_off(info, tmpul);
			return 0;

		default:
		return n_tty_ioctl(tty, file, cmd, arg); // FIXME This is wrongwrongwrong, but how do I make it right?
		//			return -EINVAL;
	}

	return -EINVAL;
}

static void arcserial_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct aa3_serial *info = (struct aa3_serial *)tty->driver_data;

	//printk("AA3 serial: in arcserial_set_termios\n");
	if (tty->termios->c_cflag == old_termios->c_cflag)
		return;

	change_speed(info, CONSOLE_BAUD_RATE); 

	if ((old_termios->c_cflag & CRTSCTS) &&
			!(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped = 0;
		arcserial_start(tty);
	}
}

/*
 * ------------------------------------------------------------
 * arcserial_close()
 * 
 * This routine is called when the serial port gets closed.	 First, we
 * wait for the last remaining data to be sent. Then, we unlink its
 * async structure from the interrupt chain if necessary, and we free
 * that IRQ if nothing is left in the chain.
 * ------------------------------------------------------------
 */
static void arcserial_close(struct tty_struct *tty, struct file * filp)
{
	struct aa3_serial * info = (struct aa3_serial *)tty->driver_data;
	aa3_uart *uart = uart_addr[info->line];
	unsigned long flags;

	if (!info || serial_paranoia_check(info, tty->device, "aa3_close"))
		return;
	
	save_flags(flags); cli();
	
	if (tty_hung_up_p(filp)) {
		restore_flags(flags);
		return;
	}
	
	if ((tty->count == 1) && (info->count != 1)) {
		/*
		 * Uh, oh. tty->count is 1, which means that the tty
		 * structure will be freed. Info->count should always
		 * be one in these conditions. If it's greater than
		 * one, we've got real problems, since it means the
		 * serial port won't be shutdown.
		 */
		printk("arcserial_close: bad serial port count; tty->count is 1, "
					 "info->count is %d\n", info->count);
		info->count = 1;
	}
	if (--info->count < 0) {
		printk("arcserial_close: bad serial port count for ttyS%d: %d\n",
					 info->line, info->count);
		info->count = 0;
	}
	if (info->count) {
		restore_flags(flags);
		return;
	}
	info->flags |= ASYNC_CLOSING;
	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	if (info->flags & ASYNC_NORMAL_ACTIVE)
		info->normal_termios = *tty->termios;
	if (info->flags & ASYNC_CALLOUT_ACTIVE)
		info->callout_termios = *tty->termios;
	/*
	 * Now we wait for the transmit buffer to clear; and we notify 
	 * the line discipline to only process XON/XOFF characters.
	 */
	tty->closing = 1;
	if (info->closing_wait != ASYNC_CLOSING_WAIT_NONE)
		tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * At this point we stop accepting input. To do this, we
	 * disable the receive line status interrupts, and tell the
	 * interrupt driver to stop checking the data ready bit in the
	 * line status register.
	 */

	uart->status &= ~UART_RXENB;

	shutdown(info);
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
	tty_ldisc_flush(tty);
	tty->closing = 0;
	info->event = 0;
	info->tty = 0;
#if 0 // member ldisc removed in 2.4.29
	if (tty->ldisc.num != ldiscs[N_TTY].num) {
		if (tty->ldisc.close)
			(tty->ldisc.close)(tty);
		tty->ldisc = ldiscs[N_TTY];
		tty->termios->c_line = N_TTY;
		if (tty->ldisc.open)
			(tty->ldisc.open)(tty);
	}
#endif	
	if (info->blocked_open) {
		if (info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(info->close_delay);
		}
		wake_up_interruptible(&info->open_wait);
	}
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|
			 ASYNC_CLOSING);
	wake_up_interruptible(&info->close_wait);
	restore_flags(flags);
}

/*
 * arcserial_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
void arcserial_hangup(struct tty_struct *tty)
{
	struct aa3_serial * info = (struct aa3_serial *)tty->driver_data;
	
	if (serial_paranoia_check(info, tty->device, "arcserial_hangup"))
		return;
	
	arcserial_flush_buffer(tty);
	shutdown(info);
	info->event = 0;
	info->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}

/*
 *	arcserial_open and friends 
 */
static int block_til_ready(struct tty_struct *tty, struct file * filp,
				 struct aa3_serial *info)
{
	DECLARE_WAITQUEUE(wait, current);
	int		retval;
	int		do_clocal = 0;

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */

	if (info->flags & ASYNC_CLOSING) {
		interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
		if (info->flags & ASYNC_HUP_NOTIFY)
			return -EAGAIN;
		else
			return -ERESTARTSYS;
#else
		return -EAGAIN;
#endif
	}

	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */
	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if (info->flags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
				(info->flags & ASYNC_SESSION_LOCKOUT) &&
				(info->session != current->session))
				return -EBUSY;
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
				(info->flags & ASYNC_PGRP_LOCKOUT) &&
				(info->pgrp != current->pgrp))
				return -EBUSY;
		info->flags |= ASYNC_CALLOUT_ACTIVE;
		return 0;
	}
	
	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) ||
			(tty->flags & (1 << TTY_IO_ERROR))) {
		if (info->flags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;
		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}

	if (info->flags & ASYNC_CALLOUT_ACTIVE) {
		if (info->normal_termios.c_cflag & CLOCAL)
			do_clocal = 1;
	} else {
		if (tty->termios->c_cflag & CLOCAL)
			do_clocal = 1;
	}
	
	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout). While we are in
	 * this loop, info->count is dropped by one, so that
	 * arcserial_close() knows when to free things. We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);

	info->count--;
	info->blocked_open++;
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		if (tty_hung_up_p(filp) ||
				!(info->flags & ASYNC_INITIALIZED)) {
#ifdef SERIAL_DO_RESTART
			if (info->flags & ASYNC_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;	
#else
			retval = -EAGAIN;
#endif
			break;
		}
		if (!(info->flags & ASYNC_CALLOUT_ACTIVE) &&
				!(info->flags & ASYNC_CLOSING) && do_clocal)
			break;
								if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			break;
		}
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);
	if (!tty_hung_up_p(filp))
		info->count++;
	info->blocked_open--;

	if (retval)
		return retval;
	info->flags |= ASYNC_NORMAL_ACTIVE;
	return 0;
} 

int arcserial_open(struct tty_struct *tty, struct file * filp)
{
		struct aa3_serial *info;
		int retval, line;
	
	line = MINOR(tty->device) - tty->driver.minor_start;
	if ((line < 0) || (line >= NR_PORTS)) {
		printk("AA3 serial driver: check your lines\n");
		return -ENODEV;
	}

	info = &aa3_info[line];
	//	printk("AA3 serial driver: opening tty %d, UART at 0x%x\n", line, info->port);

	if (serial_paranoia_check(info, tty->device, "aa3_open"))
	{
		//		printk("AA3 serial driver: no device, tty %d\n", line);
		return -ENODEV;
	}

	info->count++;
	tty->driver_data = info;
	info->tty = tty;

	/*
	 * Start up serial port
	 */
	retval = startup(info);
	if (retval)
		return retval;

	retval = block_til_ready(tty, filp, info); 
	if (retval) 
		return retval;

	if ((info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = info->normal_termios;
		else 
			*tty->termios = info->callout_termios;
		change_speed(info, CONSOLE_BAUD_RATE); 
	}
	info->session = current->session;
	info->pgrp = current->pgrp;

	//	printk("AA3 serial driver: opened tty %d, UART at 0x%x\n", line, info->port);
	return 0;
}

static int __init aa3_serial_init(void)
{
	struct aa3_serial *info; 
	int flags,i; 
	unsigned long tmp;

	memset((void*) &IRQ_ports, 0, sizeof(IRQ_ports));
		
	/* Setup base handler, and timer table. */
	init_bh(SERIAL_BH, do_serial_bh);

	/* Initialize the tty_driver structure */
	memset(&serial_driver, 0, sizeof(struct tty_driver));
	serial_driver.magic = TTY_DRIVER_MAGIC;
	serial_driver.name = "ttyS";
	serial_driver.major = TTY_MAJOR;
	serial_driver.minor_start = 64;
	serial_driver.num = NR_PORTS;
	serial_driver.type = TTY_DRIVER_TYPE_SERIAL;
	serial_driver.subtype = SERIAL_TYPE_NORMAL;
	serial_driver.init_termios = tty_std_termios;

	serial_driver.init_termios.c_cflag =
		B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	serial_driver.flags = TTY_DRIVER_REAL_RAW;
	serial_driver.refcount = &serial_refcount;
	serial_driver.table = serial_table;
	serial_driver.termios = serial_termios;
	serial_driver.termios_locked = serial_termios_locked;

	serial_driver.open = arcserial_open;
	serial_driver.close = arcserial_close;
	serial_driver.write = arcserial_write;
	serial_driver.flush_chars = arcserial_flush_chars;
	serial_driver.write_room = arcserial_write_room;
	serial_driver.chars_in_buffer = arcserial_chars_in_buffer;
	serial_driver.flush_buffer = arcserial_flush_buffer;
	serial_driver.ioctl = arcserial_ioctl;
	serial_driver.throttle = arcserial_throttle;
	serial_driver.unthrottle = arcserial_unthrottle;
	serial_driver.set_termios = arcserial_set_termios;
	serial_driver.stop = arcserial_stop;
	serial_driver.start = arcserial_start;
	serial_driver.hangup = arcserial_hangup;
//	serial_driver.set_ldisc = arcserial_set_ldisc;
	serial_driver.driver_name = "serial";

	/*
	 * The callout device is just like normal device except for
	 * major number and the subtype code.
	 */
	callout_driver = serial_driver;
	callout_driver.name = "cua";
	callout_driver.major = TTYAUX_MAJOR;
	callout_driver.subtype = SERIAL_TYPE_CALLOUT;

	if (tty_register_driver(&serial_driver)) {
		printk("%s: Couldn't register serial driver\n", __FUNCTION__);
		return(-EBUSY);
	}
	if (tty_register_driver(&callout_driver)) {
		printk("%s: Couldn't register callout driver\n", __FUNCTION__);
		return(-EBUSY);
	}
	
	save_flags(flags); cli();

	for(i=0;i<NR_PORTS;i++) {
		
			info = &aa3_info[i];
			info->magic = SERIAL_MAGIC;
			info->port = (unsigned int) uart_addr[i];
			info->tty = 0;
			info->irq = uart_irqs[i];
			info->custom_divisor = 16;
			info->close_delay = 50;
			info->closing_wait = 3000;
			info->x_char = 0;
			info->event = 0;
			info->count = 0;
			info->blocked_open = 0;
			info->tqueue.routine = do_softint;
			info->tqueue.data = info;
			info->tqueue_hangup.routine = do_serial_hangup;
			info->tqueue_hangup.data = info;
			info->callout_termios = callout_driver.init_termios;
			info->normal_termios = serial_driver.init_termios;
			init_waitqueue_head(&info->open_wait);
			init_waitqueue_head(&info->close_wait);
			info->line = i;
			//			info->is_cons = ((i == 0) ? 1 : 0); 
			info->is_cons = ((i == 1) ? 1 : 0); 
			
			printk("%s%d (irq = %d, address = 0x%x)", serial_driver.name, info->line, 
			 info->irq, info->port);
			printk(" is a builtin ARC UART\n");
			
			IRQ_ports[info->irq] = info;			 

			if (i < MAX_NAMED_UART)
			{
				if (request_irq(uart_irqs[i], aa3_serial_interrupt, IRQ_FLG_STD,
												uart_names[i], NULL))
					panic("Unable to attach arc serial interrupt\n");
			}
			else
			{
				if (request_irq(uart_irqs[i], aa3_serial_interrupt, IRQ_FLG_STD,
												"UART", NULL))
					panic("Unable to attach arc serial interrupt\n");
			}
	    /* Make the interrupt edge triggered to try and eliminate spurious interrupts */
	    tmp = read_new_aux_reg(AUX_ITRIGGER);
	    tmp |= (1<<uart_irqs[i]);
	    write_new_aux_reg(AUX_ITRIGGER,tmp);
#ifdef CONFIG_ARC700_SERIAL
	    tmp = read_new_aux_reg(AUX_IENABLE);
	    tmp |= (1<<uart_irqs[i]);
	    write_new_aux_reg(AUX_IENABLE,tmp);
#endif
	}
	restore_flags(flags);

#ifdef CONFIG_SKY_IR
	if(check_mem_region(IRBaseIoAddress, IRBaseIoAddress_len)){
		printk("ERROR: IR I/O memory already in use.");
		return -1;
	}

	request_mem_region(IRBaseIoAddress, IRBaseIoAddress_len, "IR");
	
	ir_mem = ioremap(IRBaseIoAddress, IRBaseIoAddress_len);

	// register isr
	 if(request_irq(IR_INT_NUM, sky_ir_interrupt, 0 , "SKY IR", NULL)){
                info("request_irq function error");
                return;
        }else{
                info("IRQ=%d", IR_INT_NUM);
        } 

	IR_INT_ENABLE;
	info("IR3_CON=0x%08x", IR3CON);

	IR_CAN_INIT;
	IR_CBN_INIT;
	IR_CCN_INIT;
#endif

	return 0;
}

/*
 * register_serial and unregister_serial allows for serial ports to be
 * configured at run-time, to support PCMCIA modems.
 */
int register_serial(struct serial_struct *req)
{
	return -1;
}

void unregister_serial(int line)
{
	return;
}

module_init(aa3_serial_init);

int aa3_console_setup(struct console *cp, char *arg)
{
	aa3_uart *uart;
	//		aa3_uart *uart = &uart_addr[0];
	int baud_h, baud_l;
	int i;
	/* The default clock speed is set at config-time in CONFIG_ARC700_CLK,
	   then is held in arc700_clkfreq so it can be altered by a cmdline arg */
	int arc_console_baud = (arc700_clkfreq/(CONSOLE_BAUD_RATE*4)) - 1;

	for (i = 0; i < NR_PORTS; i++)
	{ 
		uart = uart_addr[i];
		printk("Resetting ttyS%d\n", i);
		baud_l = (arc_console_baud & 0xff);
		baud_h = (arc_console_baud & 0xff00) >> 8;

		/* 1. set RXENB (bit 2) to receive interrupts on data receive.
		 *
		 * 2. store value of the baud rate in BL and BH.
		 */
		uart->status = UART_RXENB;
		uart->baudl = baud_l;
		uart->baudh = baud_h;
	}
	console_inited = 1 ;

	return 0; /* successful initialization */
}


static kdev_t aa3_console_device(struct console *c)
{
	return MKDEV(TTY_MAJOR, 64 + c->index);
}

void aa3_console_write(struct console *cp, const char *p, unsigned len)
{
	if (!console_inited)
		aa3_console_setup(cp, NULL);
	while (len-- > 0) {

		if (*p == '\n')
			arcserial_put_char('\r');
		arcserial_put_char(*p++);
	}
}

struct console arc_aa3_console = {
	.name		 = "ttyS",
	.write	 = aa3_console_write,
	.read		 = NULL, /*aa3_console_read,*/
	.device	 = aa3_console_device,
	.unblank = NULL,
	.setup	 = aa3_console_setup,
	.flags	 = CON_PRINTBUFFER,
	.index	 = -1,
	.cflag	 = 0,
	.next		 = NULL
};

void __init aa3_console_init(void)
{
	register_console(&arc_aa3_console);
}
