/*
 *	ARCDog	A Hardware Watchdog Device for ARC Processors
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *	
 *	based on softdog.c via indydog.c
 */
 
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/arcregs.h>

extern unsigned long arc700_clkfreq;

static unsigned long arcdog_alive;
static int expect_close = 0;

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
#else
static int nowayout = 0;
#endif

MODULE_PARM(nowayout,"i");
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");

static inline void arcdog_ping(void)
{
  /* Set counter to 0 */
  write_new_aux_reg(ARC_REG_TIMER1_CNT, 0);
}

/*
 *	Allow only one person to hold it open
 */
static int arcdog_open(struct inode *inode, struct file *file)
{
	u32 mc_ctrl0;
	u32 limit;
	
	if (test_and_set_bit(0,&arcdog_alive))
		return -EBUSY;

	if (nowayout) {
		MOD_INC_USE_COUNT;
	}

	limit = (arc700_clkfreq * CONFIG_ARCDOG_TIMEOUT); /* Max timeout is ~8 secs at 500MHz, ~45s at 90MHz */
  write_new_aux_reg(ARC_REG_TIMER1_CTRL, 0);
	/* Activate timer */
  write_new_aux_reg(ARC_REG_TIMER1_LIMIT, limit);
  /* Set control with IE and NH bits set */
  write_new_aux_reg(ARC_REG_TIMER1_CTRL, 6); /* Watchdog + Not Halted modes */
	
	arcdog_ping();
			
	arcdog_alive = 1;
	printk(KERN_INFO "Started watchdog timer.\n");
	
	return 0;
}

static int arcdog_release(struct inode *inode, struct file *file)
{
	/* Shut off the timer.
	 * Lock it in if it's a module and we set nowayout. */
	lock_kernel();
	if (expect_close) {
		write_new_aux_reg(ARC_REG_TIMER1_CTRL, 0);
		printk(KERN_INFO "Stopped watchdog timer.\n");
	} else
		printk(KERN_CRIT "Watchdog device closed unexpectedly - will not stop!\n");
	clear_bit(0, &arcdog_alive);
	unlock_kernel();

	return 0;
}

static ssize_t arcdog_write(struct file *file, const char *data, size_t len, loff_t *ppos)
{
	/*  Can't seek (pwrite) on this device  */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	/*
	 *	Refresh the timer.
	 */
	if (len) {
		if (!nowayout) {
			size_t i;

			/* In case it was set long ago */
			expect_close = 0;
 
			for (i = 0; i != len; i++) {
				char c;
				if (get_user(c, data + i))
					return -EFAULT;
				if (c == 'V')
					expect_close = 1;
			}
		}
		arcdog_ping();
		return 1;
	}
	return 0;
}

static int arcdog_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	static struct watchdog_info ident = {
		options: WDIOF_MAGICCLOSE,
		identity: "Hardware Watchdog for ARC",
	};
	switch (cmd) {
		default:
			return -ENOIOCTLCMD;
		case WDIOC_GETSUPPORT:
			if(copy_to_user((struct watchdog_info *)arg, &ident, sizeof(ident)))
				return -EFAULT;
			return 0;
		case WDIOC_GETSTATUS:
		case WDIOC_GETBOOTSTATUS:
			return put_user(0,(int *)arg);
		case WDIOC_KEEPALIVE:
			arcdog_ping();
			return 0;
	}
}

static struct file_operations arcdog_fops = {
	owner:		THIS_MODULE,
	write:		arcdog_write,
	ioctl:		arcdog_ioctl,
	open:		arcdog_open,
	release:	arcdog_release,
};

static struct miscdevice arcdog_miscdev = {
	minor:		WATCHDOG_MINOR,
	name:		"watchdog",
	fops:		&arcdog_fops,
};

static const char banner[] __initdata = KERN_INFO "Hardware Watchdog Timer for ARC processors\n";

static int __init watchdog_init(void)
{
	int ret = misc_register(&arcdog_miscdev);

	if (ret)
		return ret;

	printk(banner);

	return 0;
}

static void __exit watchdog_exit(void)
{
	misc_deregister(&arcdog_miscdev);
}

module_init(watchdog_init);
module_exit(watchdog_exit);
MODULE_LICENSE("GPL");
