/******************************************************************************
 * Copyright Codito Technologies (www.codito.com) Oct 01, 2004
 * 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *****************************************************************************/

/*
 * arch/arcnommu/kernel/time.c
 *
 * Initially added to get the timer interrupt up
 *
 * Authors: Amit Shah, Rahul Trivedi
 */

#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/init.h>
#include <linux/timex.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/arcregs.h>


extern rwlock_t xtime_lock;
extern unsigned long wall_jiffies;
extern unsigned long arc700_clkfreq;

#define USECS_PER_JIFFY (1000000/HZ)

extern void setup_timer(void)
{
	/*
	 * Initializing timer 0:
	 * steps: 1. put the limit[1] in the limit register 0x23
	 *        2. set the control bits in the control reg. 0x22
	 *           bit 0: enable interrupts
	 *           bit 1: only count when proc. is not in halt mode.
	 *        3. set the initial timer count value, reg. 0x21
	 *
	 * [1] limit: the timer counts from the initial count value to
	 * the value in the limit register before interrupting. We
	 * want to set this to 10ms. The clock frequency is in
	 * CONFIG_ARC700_CLK. (clk freq / 10e-3)
	 */

//  write_new_aux_reg(ARC_REG_TIMER0_LIMIT, (arc700_clkfreq/100));
  write_new_aux_reg(ARC_REG_TIMER0_LIMIT, (arc700_clkfreq/HZ));
  /* Set counter to 0 */
  write_new_aux_reg(ARC_REG_TIMER0_CNT, 0);
  /* Set control with IE and NH bits set */
  write_new_aux_reg(ARC_REG_TIMER0_CTRL, 3);

  return;
}


void timer_handler(int irq, void *dev_id, struct pt_regs *fp)
{
	do_timer(fp);
	/* ack the timer interrupt */
	write_new_aux_reg(ARC_REG_TIMER0_CTRL, 3);
}

/* Some functions here are taken from arch/armnommu/kernel/ */

static int dummy_set_rtc(void)
{
	return 0;
}

/*
 * hook for setting the RTC's idea of the current time.
 */
int (*set_rtc)(void) = dummy_set_rtc;

static unsigned long dummy_gettimeoffset(void)
{
	return 0;
}

/*
 * hook for getting the time offset.  Note that it is
 * always called with interrupts disabled.
 */
unsigned long (*gettimeoffset)(void) = dummy_gettimeoffset;

void do_gettimeofday(struct timeval *tv)
{
	unsigned long flags;
	unsigned long usec, sec;

	read_lock_irqsave(&xtime_lock, flags);
	usec = gettimeoffset();
	{
		unsigned long lost = jiffies - wall_jiffies;

		if (lost)
			usec += lost * USECS_PER_JIFFY;
	}
	sec = xtime.tv_sec;
	usec += xtime.tv_usec;
	read_unlock_irqrestore(&xtime_lock, flags);

	/* usec may have gone up a lot: be safe */
	while (usec >= 1000000) {
		usec -= 1000000;
		sec++;
	}

	tv->tv_sec = sec;
	tv->tv_usec = usec;
}

void do_settimeofday(struct timeval *tv)
{
	write_lock_irq(&xtime_lock);
	/* This is revolting. We need to set the xtime.tv_usec
	 * correctly. However, the value in this location is
	 * is value at the last tick.
	 * Discover what correction gettimeofday
	 * would have done, and then undo it!
	 */
	tv->tv_usec -= gettimeoffset();
	tv->tv_usec -= (jiffies - wall_jiffies) * USECS_PER_JIFFY;

	while (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_adjust = 0;		/* stop active adjtime() */
	time_status |= STA_UNSYNC;
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;
	write_unlock_irq(&xtime_lock);
}

static struct irq_node timer_intr = {timer_handler, IRQ_FLG_LOCK, 0, NULL,
				     "Timer", NULL };

void __init time_init(void)
{
	xtime.tv_usec = 0;
	xtime.tv_sec = 0;
//	xtime.tv_sec = 86400;
	xtime.tv_sec = 1167580800 + 28800;	//2007/01/01

/* insert the timer interrupt ISR into the irq list */	
	setup_arc_irq(TIMER_INT, &timer_intr);
	setup_timer();
}
